#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2020 Michael J. Hayford
"""Utilities including Singleton metaclass

.. Created on Wed Sep 16 21:42:49 2020

.. codeauthor: Michael J. Hayford
"""


class Counter(dict):
    """A dict that initializes a missing key's value to 0.

    Example:
        track_changes = Counter()
        track_changes['something happened'] += 1
        track_changes['something not found'] += 1
        """

    def __missing__(self, key):
        return 0


class Singleton(type):
    """A metaclass implementation for the Singleton pattern.

    Example:

        class JustOne(metaclass=Singleton):
            pass
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = (super(Singleton, cls).
                                   __call__(*args, **kwargs))
        return cls._instances[cls]


def rgb2mpl(rgb):
    """ convert 8 bit RGB data to 0 to 1 range for mpl """
    if len(rgb) == 3:
        return [rgb[0]/255., rgb[1]/255., rgb[2]/255., 1.0]
    elif len(rgb) == 4:
        return [rgb[0]/255., rgb[1]/255., rgb[2]/255., rgb[3]/255.]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2017 Michael J. Hayford
""" Support for the Ohara Glass catalog

.. codeauthor: Michael J. Hayford
"""
import logging
from .util import Singleton

from math import sqrt

from . import glass


class OharaCatalog(glass.GlassCatalog, metaclass=Singleton):
    #    data_header = 1
    #    data_start = 2
    #    num_glasses = 134
    #    name_col_offset = 1
    #    coef_col_offset = 60
    #    index_col_offset = 4
    nline_str = {'nC': 'nC',
                 'nd': 'nd',
                 'ne': 'ne',
                 'nF': 'nF',
                 'ng': 'ng',
                 'nh': 'nh',
                 'ni': 'ni'}

    def __init__(self, fname='OHARA.xlsx'):
        super().__init__('Ohara', fname, 'Glass ', 'A1', 'n2325')

    def create_glass(self, gname, gcat):
        return OharaGlass(gname)


class OharaGlass(glass.Glass):
    catalog = OharaCatalog()

    def __init__(self, gname, catalog=None):
        if catalog is not None:
            self.catalog = catalog
        super().__init__(gname)

    def glass_code(self):
        return super().glass_code('nd', 'νd')

    def calc_rindex(self, wv_nm):
        wv = 0.001*wv_nm
        wv2 = wv*wv
        coefs = self.catalog.glass_coefs(self.gindex)
        n2 = 1 + coefs[0]*wv2/(wv2 - coefs[3])
        n2 += coefs[1]*wv2/(wv2 - coefs[4])
        n2 += coefs[2]*wv2/(wv2 - coefs[5])
        return sqrt(n2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2020 Michael J. Hayford
"""

.. Created on Thu Sep 24 22:09:48 2020

.. codeauthor: Michael J. Hayford
"""
import numpy as np

from matplotlib.figure import Figure
from matplotlib.patches import Polygon

from . import glass as cat_glass
from . import glassfactory as gf
from . import glasspolygons as gp
from . import util


class GlassMapDB():
    """ Simple model to support Model/View architecture for Glass map views """

    def __init__(self, *args):
        # if no arguments, use the default set of catalogs
        if len(args) == 0:
            args = (gf._catalog_list,)

        self.catalogs = []
        for arg in args:
            if isinstance(arg, dict):
                for cat_name, catalog in arg.items():
                    self.catalogs.append((catalog, cat_name))
            if isinstance(arg, list):
                if isinstance(arg[0], str):
                    # treat as a list of catalog names
                    for cat_name in arg:
                        catalog = gf.get_glass_catalog(cat_name)
                        self.catalogs.append((catalog, cat_name))
                else:
                    # treat as glass instances, wrap them in a GlassMapSet
                    catalog = GlassMapSet(arg)
                    self.catalogs.append((catalog, 'System'))

    def get_data_at(self, i, **kwargs):
        catalog, cat_name = self.catalogs[i]
        return catalog.glass_map_data(cat_name=cat_name, **kwargs)

    def get_data_set_label_at(self, i):
        return self.catalogs[i][1]


class GlassMapSet():
    """Set of glass instances to be displayed in a GlassMapFigure. """

    def __init__(self, glass_list):
        """The main requirement is the glass instance must respond to
        rindex() api
        """
        self.glass_list = glass_list

    def glass_map_data(self, wvl='d', **kwargs):
        return calc_glass_map_arrays(self.glass_list, wvl, 'F', 'C', **kwargs)


def calc_glass_map_arrays(glass_list, d_str, F_str, C_str, **kwargs):
    """ return index and dispersion data arrays for input spectral range

    Args:
        nd_str (str): central wavelength string
        nf_str (str): blue end wavelength string
        nc_str (str): red end wavelength string

    Returns:
        index, V-number, partial dispersion, Buchdahl coefficients, and
        glass names
    """

    nd = np.array([g.rindex(d_str) for g in glass_list])
    nF = np.array([g.rindex(F_str) for g in glass_list])
    nC = np.array([g.rindex(C_str) for g in glass_list])

    vd, PCd = cat_glass.calc_glass_constants(nd, nF, nC)

    nd, coefs = cat_glass.calc_buchdahl_coords(
        nd, nF, nC, wlns=(d_str, F_str, C_str), **kwargs)

    names = [g.name()+'/'+g.catalog_name() for g in glass_list]
    return nd, vd, PCd, coefs[0], coefs[1], names


class GlassMapFigure(Figure):
    dsc = [(56/255, 142/255, 142/255),  # sgi teal
           (133/255, 133/255, 133/255),  # grey 52
           (113/255, 113/255, 198/255),  # sgi slateblue
           (102/255, 205/255, 0),  # chartreuse 3
           (255/255, 114/255, 86/255),  # coral 1
           (255/255, 165/255, 0/255)]  # orange 1

    def __init__(self, glass_db, db_display, plot_display_type,
                 width=5, height=4, refresh_gui=None, **kwargs):
        super().__init__(figsize=(width, height), **kwargs)
        self.refresh_gui = refresh_gui
        self.glass_db = glass_db
        self.db_display = db_display
        self.plot_display_type = plot_display_type
        self.needsClear = True
        self.pick_list = []

        self.update_data()

    def get_display_label(self):
        return self.plot_display_type

    def refresh(self, **kwargs):
        """Call update_data() followed by plot(), return self.

        Args:
            kwargs: keyword arguments are passed to update_data

        Returns:
            self (class Figure) so scripting envs will auto display results
        """
        self.update_data(**kwargs)
        self.plot()
        return self

    def update_data(self, **kwargs):
        self.rawData = []
        ctyp = ("disp_coefs"
                if self.plot_display_type == "Buchdahl Dispersion Coefficients"
                else None)
        for i, display in enumerate(self.db_display):
            gmap_data = self.glass_db.get_data_at(i, ctype=ctyp)
            n, v, p, coefs0, coefs1, lbl = gmap_data
            dsLabel = self.glass_db.get_data_set_label_at(i)
            self.rawData.append([dsLabel, (n, v, p, coefs0, coefs1, lbl)])
        return self

    def draw_axes(self):
        self.ax.grid(True)
        if hasattr(self, 'header'):
            self.ax.set_title(self.header, pad=10.0, fontsize=18)
        if hasattr(self, 'x_label'):
            self.ax.set_xlabel(self.x_label)
        if hasattr(self, 'y_label'):
            self.ax.set_ylabel(self.y_label)

    def plot(self):
        try:
            self.ax.cla()
        except AttributeError:
            self.ax = self.add_subplot(1, 1, 1)

        if self.plot_display_type == "Refractive Index":
            self.x_label = r'$\mathrm{V_d}$'
            self.y_label = r'$\mathrm{n_d}$'
            xi = 1
            yi = 0
            self.draw_glass_polygons()
        elif self.plot_display_type == "Partial Dispersion":
            self.x_label = r'$\mathrm{V_d}$'
            self.y_label = r'$\mathrm{P_{C-d}}$'
            xi = 1
            yi = 2
        elif self.plot_display_type == "Buchdahl Coefficients":
            self.x_label = r'$\mathrm{\nu_2}$'
            self.y_label = r'$\mathrm{\nu_1}$'
            xi = 4
            yi = 3
        elif self.plot_display_type == "Buchdahl Dispersion Coefficients":
            self.x_label = r'$\mathrm{\eta_2}$'
            self.y_label = r'$\mathrm{\eta_1}$'
            xi = 4
            yi = 3
        self.ax.set_title(self.get_display_label())
        for i, display in enumerate(self.db_display):
            self.ax.plot(self.rawData[i][1][xi], self.rawData[i][1][yi],
                         linestyle='None', marker='o', markersize=5,
                         alpha=0.75, gid=i, picker=5,
                         # color=self.dsc[i], alpha=0.75, gid=i, picker=5,
                         label=self.rawData[i][0], visible=display)

        self.canvas.mpl_connect('pick_event', self.on_pick)
        self.canvas.mpl_connect('button_press_event', self.on_press)

        if xi == 1:
            self.ax.invert_xaxis()
        self.draw_axes()
        self.ax.legend()
        self.canvas.draw()
        return self

    def draw_glass_polygons(self):
        for glass, poly in gp.polygons.items():
            rgb = gp.rgb[glass]
            p = Polygon(poly, closed=True, fc=util.rgb2mpl(rgb), ec='black',
                        linewidth=1.0)
            self.ax.add_artist(p)

    def clear_pick_table(self):
        self.pick_list = []
        self.needsClear = False

    def on_press(self, event):
        if self.needsClear:
            # If needsClear is still set, there have been no pick events so
            #  this is a click in an empty region of the plot.
            #  Clear the pick table
            self.clear_pick_table()
        else:
            # on_press event happens after on_pick events. Set needsClear for
            #  next on_pick, i.e. a new selection, to handle
            self.needsClear = True
        if self.refresh_gui is not None:
            self.refresh_gui()

    def on_pick(self, event):
        if self.needsClear:
            self.clear_pick_table()
        line = event.artist
        id = line.get_gid()
        if self.db_display[id]:
            ind = event.ind
            dsLabel = self.rawData[id][0]
            n, v, p, coef0, coef1, lbl = self.rawData[id][1]
            for k in ind:
                glass = (dsLabel, lbl[k], n[k], v[k], p[k])
                self.pick_list.append(glass)

    def updateVisibility(self, indx, state):
        self.ax.lines[indx].set_visible(state)
        self.canvas.draw()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" unit test for Schott optical glass catalog

.. codeauthor: Michael J. Hayford
"""

import unittest
import opticalglass.schott as s


class SchottTestCase(unittest.TestCase):
    catalog = s.SchottCatalog()

    def compare_indices(self, glass, tol=5e-6):
        nC = glass.rindex(656.27)
        nd = glass.rindex(587.56)
        ne = glass.rindex(546.07)
        nF = glass.rindex(486.13)
        ng = glass.rindex(435.84)
        nh = glass.rindex(404.66)
        nI = glass.rindex(365.01)
        indxC = glass.glass_data()[self.catalog.data_index('  nC')]
        indxd = glass.glass_data()[self.catalog.data_index('  nd')]
        indxe = glass.glass_data()[self.catalog.data_index('  ne')]
        indxF = glass.glass_data()[self.catalog.data_index('  nF')]
        indxg = glass.glass_data()[self.catalog.data_index('  ng')]
        indxh = glass.glass_data()[self.catalog.data_index('  nh')]
        indxI = glass.glass_data()[self.catalog.data_index('  ni')]
        self.assertAlmostEqual(nC, indxC, delta=tol)
        self.assertAlmostEqual(nd, indxd, delta=tol)
        self.assertAlmostEqual(ne, indxe, delta=tol)
        self.assertAlmostEqual(nF, indxF, delta=tol)
        self.assertAlmostEqual(ng, indxg, delta=tol)
        self.assertAlmostEqual(nh, indxh, delta=tol)
        self.assertAlmostEqual(nI, indxI, delta=tol)

    def test_schott_catalog_glass_index(self):
        f2 = self.catalog.glass_index('F2')
        self.assertEqual(f2, 0)
        nbk7 = self.catalog.glass_index('N-BK7')
        self.assertEqual(nbk7, 25)
        sf6ht = self.catalog.glass_index('SF6HT')
        self.assertEqual(sf6ht, 122)

    def test_schott_catalog_data_index(self):
        nd = self.catalog.data_index('nd')
        self.assertEqual(nd, 1)
        vd = self.catalog.data_index('vd')
        self.assertEqual(vd, 3)
        B1 = self.catalog.data_index('B1')
        self.assertEqual(B1, 6)
        glasscode = self.catalog.data_index('Glascode')
        self.assertEqual(glasscode, 158)
        date = self.catalog.data_index('Date')
        self.assertEqual(date, 160)

    def test_schott_glass_f2(self):
        f2 = s.SchottGlass('F2')
        self.assertIsNotNone(f2.gindex)
        self.assertEqual(f2.name(), 'F2')
        self.compare_indices(f2)

    def test_schott_glass_nbk7(self):
        nbk7 = s.SchottGlass('N-BK7')
        self.assertIsNotNone(nbk7.gindex)
        self.assertEqual(nbk7.name(), 'N-BK7')
        self.compare_indices(nbk7)

    def test_schott_glass_sf6ht(self):
        sf6ht = s.SchottGlass('SF6HT')
        self.assertIsNotNone(sf6ht.gindex)
        self.assertEqual(sf6ht.name(), 'SF6HT')
        self.compare_indices(sf6ht)


if __name__ == '__main__':
    unittest.main(verbosity=2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Support for Glass catalog exception handling

.. codeauthor: Michael J. Hayford
"""


class GlassError(Exception):
    """ Exception raised when interrogating glass database """


class GlassCatalogNotFoundError(GlassError):
    """ Exception raised when glass catalog name not found """

    def __init__(self, catalog):
        self.catalog = catalog


class GlassNotFoundError(GlassError):
    """ Exception raised when glass name not found """

    def __init__(self, catalog, name):
        self.catalog = catalog
        self.name = name


class GlassDataNotFoundError(GlassError):
    """ Exception raised when glass data item not found """

    def __init__(self, catalog, data):
        self.catalog = catalog
        self.data = data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" desktop application for viewing glass catalog data

.. Created on Wed Jan  3 12:50:03 2018

.. codeauthor: Michael J. Hayford
"""

import sys

from PyQt5.QtWidgets import (QApplication, QMainWindow, QWidget, QHBoxLayout,
                             QVBoxLayout, QSizePolicy, QGroupBox, QCheckBox,
                             QRadioButton, QTableView)
from PyQt5.QtCore import Qt, QAbstractTableModel, QModelIndex, QMimeData
from PyQt5.QtGui import QDrag

from matplotlib.backends.backend_qt5agg \
     import (FigureCanvasQTAgg as FigureCanvas,
             NavigationToolbar2QT as NavigationToolbar)

from opticalglass.glassmap import GlassMapFigure, GlassMapDB


def init_UI(gui_parent, fig):
    main_widget = QWidget()
    layout = QHBoxLayout(main_widget)

    gm = PlotCanvas(gui_parent, fig)
    layout.addWidget(gm)

    gui_parent.addToolBar(Qt.BottomToolBarArea,
                          NavigationToolbar(gm, gui_parent))

    rightBar = QVBoxLayout()
    layout.addLayout(rightBar)

    plotTypeGroup = createPlotTypeBox(gui_parent, fig)
    rightBar.addWidget(plotTypeGroup)

    catalogGroup = createCatalogGroupBox(gui_parent, fig)
    rightBar.addWidget(catalogGroup)

    pick_model = PickModel(fig)
    gmt = PickTable(gui_parent, pick_model)
    rightBar.addWidget(gmt)

    return main_widget, gm, gmt, pick_model


def createPlotTypeBox(gui_parent, fig):
    groupBox = QGroupBox("Plot Type", gui_parent)

    index_btn = QRadioButton("Refractive Index")
    index_btn.setChecked(True)
    index_btn.toggled.connect(lambda:
                              on_plot_type_toggled(fig, index_btn))
    partial_btn = QRadioButton("Partial Dispersion")
    partial_btn.toggled.connect(lambda:
                                on_plot_type_toggled(fig, partial_btn))
    buchdahl_btn = QRadioButton("Buchdahl Coefficients")
    buchdahl_btn.toggled.connect(lambda:
                                 on_plot_type_toggled(fig, buchdahl_btn))
    buchdahl_disp_btn = QRadioButton("Buchdahl Dispersion Coefficients")
    buchdahl_disp_btn.toggled.connect(
        lambda: on_plot_type_toggled(fig, buchdahl_disp_btn))

    vbox = QVBoxLayout()
    vbox.addWidget(index_btn)
    vbox.addWidget(partial_btn)
    vbox.addWidget(buchdahl_btn)
    vbox.addWidget(buchdahl_disp_btn)

    groupBox.setLayout(vbox)

    return groupBox


def on_plot_type_toggled(fig, button):
    plot_display_type = fig.plot_display_type
    if button.text() == "Refractive Index":
        if button.isChecked() is True:
            plot_display_type = "Refractive Index"
    elif button.text() == "Partial Dispersion":
        if button.isChecked() is True:
            plot_display_type = "Partial Dispersion"
    elif button.text() == "Buchdahl Coefficients":
        if button.isChecked() is True:
            plot_display_type = "Buchdahl Coefficients"
    elif button.text() == "Buchdahl Dispersion Coefficients":
        if button.isChecked() is True:
            plot_display_type = "Buchdahl Dispersion Coefficients"

    fig.plot_display_type = plot_display_type
    fig.update_data()
    fig.plot()


def createCatalogGroupBox(gui_parent, fig):
    groupBox = QGroupBox("Glass Catalogs", gui_parent)

    check_box_list = []
    for i, gc in enumerate(fig.glass_db.catalogs):
        catalog, cat_name = gc
        checkBox = QCheckBox(cat_name)
        checkBox.setChecked(True)
        checkBox.stateChanged.connect(create_handle_checkbox(fig, i))
        check_box_list.append(checkBox)

    vbox = QVBoxLayout()
    for cb in check_box_list:
        vbox.addWidget(cb)

    groupBox.setLayout(vbox)

    return groupBox


def create_handle_checkbox(fig, cb_number):
    def handle_checkbox(state):
        checked = state == Qt.Checked
        fig.db_display[cb_number] = checked
        fig.updateVisibility(cb_number, checked)
    return handle_checkbox


class GlassMapViewer(QMainWindow):
    def __init__(self):
        super().__init__()

        self.title = 'Glass Map Viewer'
        self.setWindowTitle(self.title)

        self.left = 50
        self.top = 150
        self.width = 1100
        self.height = 650
        self.setGeometry(self.left, self.top, self.width, self.height)

        self.glass_db = GlassMapDB()
        self.db_display = [True]*len(self.glass_db.catalogs)
        self.plot_display_type = "Refractive Index"
        self.fig = GlassMapFigure(self.glass_db, self.db_display,
                                  self.plot_display_type,
                                  refresh_gui=self.refresh_gui,
                                  width=5, height=4)
        self._main, _, self.gmt, self.pick_model = init_UI(self, self.fig)
        self.setCentralWidget(self._main)

        self.fig.plot()

    def refresh_gui(self, **kwargs):
        self.pick_model.fill_table(self.fig.pick_list)


_pt_header = ["Catalog", "Glass", "Nd", "Vd", "P C,d"]
_pt_format = ["{:s}", "{:s}", "{:7.5f}", "{:5.2f}", "{:6.4f}"]


class PickTable(QTableView):
    def __init__(self, gui_parent, pick_model):
        super().__init__(gui_parent)
        self.setModel(pick_model)
        self.setAlternatingRowColors(True)
        self.setMinimumWidth(285)
        self.setMaximumWidth(335)
        self.setDragEnabled(True)
        self.pickRow = 0
        for i, w in enumerate([52, 96, 61, 48, 60]):
            self.setColumnWidth(i, w)

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        if (event.button() == Qt.LeftButton):
            drag = QDrag(self)
            mimeData = QMimeData()
            si = self.indexAt(event.pos())
            pick_row = si.row()
            pick = self.model().pick_table[pick_row]
            # comma separated list: glass_name,catalog_name
            mimeData.setText(pick[1] + ',' + pick[0])
            drag.setMimeData(mimeData)

            drag.exec_(Qt.CopyAction)


class PickModel(QAbstractTableModel):
    def __init__(self, fig):
        super().__init__()
        self.num_rows = 0
        self.pick_table = []

    def rowCount(self, index):
        return self.num_rows

    def columnCount(self, index):
        return len(_pt_header)

    def headerData(self, section, orientation, role):
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                return _pt_header[section]
            elif orientation == Qt.Vertical:
                return None
        else:
            return None

    def data(self, index, role):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            r = index.row()
            c = index.column()
            return self.pick_table[r][c]
        else:
            return None

    def fill_table(self, pick_list):
        self.pickRow = 0
        pick_table = []
        for item in pick_list:
            pick_table.append([f.format(v) for f, v in zip(_pt_format, item)])
        self.pick_table = pick_table

        if self.num_rows > 0:
            self.beginRemoveRows(QModelIndex(), 0, self.num_rows-1)
            self.removeRows(0, self.num_rows)
            self.endRemoveRows()

        self.num_rows = len(pick_table)
        if self.num_rows > 0:
            self.beginInsertRows(QModelIndex(), 0, self.num_rows-1)
            self.insertRows(0, self.num_rows)
            self.endInsertRows()


class PlotCanvas(FigureCanvas):
    def __init__(self, gui_parent, fig):
        super().__init__(fig)
        self.setParent(gui_parent)

        FigureCanvas.setSizePolicy(self,
                                   QSizePolicy.Expanding,
                                   QSizePolicy.Expanding)
        fig.plot()


def main():
    qtapp = QApplication(sys.argv)
    qtwnd = GlassMapViewer()
    qtwnd.show()
    return qtapp.exec_()


if __name__ == '__main__':
    sys.exit(main())



""" package supplying optical glass data from vendor supplied data

    The :mod:`opticalglass` package currently supports the following vendors:

        - CDGM
        - Hikari
        - Hoya
        - Ohara
        - Schott
        - Sumita

    The spreadsheets from the vendors are in the ``data`` directory.

    A set of legacy catalogs, circa 1980, is available via the
    :class:`~glass.Robb1983Catalog` class. The data used by this class is from
    the 1983 paper by Paul N. Robb and R. I. Mercado, `Calculation of
    refractive indices using Buchdahl’s chromatic coordinate
    <https://doi.org/10.1364/AO.22.001198>`_ . The catalogs include:

        - Hoya
        - Ohara
        - Schott
        - Chance
        - Corning-France

    The authors fitted a Buchdahl quadratic model to the glass data that has a
    standard deviation of 0.00002 and a maximum absolute error of 0.0001 in the
    visible spectral region.

    Fitting and modeling glass data using the Buchdahl chromatic coordinate is
    supported in the :mod:`~.buchdahl` module.
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2020 Michael J. Hayford
"""Dictionary that enables case insensitive searching

    Borrowed from https://gist.github.com/babakness/3901174 and modified to
    work with Python 3 dicts.

.. Created on Thu Sep 24 13:39:31 2020

.. codeauthor: Michael J. Hayford
"""


class CaselessDictionary(dict):
    """Dictionary that enables case insensitive searching while preserving case
    sensitivity when keys are listed, ie, via keys() or items() methods.

    Works by storing a lowercase version of the key as the new key and stores
    the original key-value pair as the key's value (values become
    dictionaries).
    """

    def __init__(self, initval={}):
        if isinstance(initval, dict):
            for key, value in iter(initval.items()):
                self.__setitem__(key, value)
        elif isinstance(initval, list):
            for (key, value) in initval:
                self.__setitem__(key, value)

    def __contains__(self, key):
        return dict.__contains__(self, key.lower())

    def __getitem__(self, key):
        return dict.__getitem__(self, key.lower())['val']

    def __setitem__(self, key, value):
        return dict.__setitem__(self, key.lower(), {'key': key, 'val': value})

    def get(self, key, default=None):
        try:
            v = dict.__getitem__(self, key.lower())
        except KeyError:
            return default
        else:
            return v['val']

    def has_key(self, key):
        if self.get(key):
            return True
        else:
            return False

    def items(self):
        return [(v['key'], v['val']) for v in iter(dict.values(self))]

    def keys(self):
        return [v['key'] for v in iter(dict.values(self))]

    def values(self):
        return [v['val'] for v in iter(dict.values(self))]
